#! /usr/bin/env python
# tag: voss@debian.org Son Feb 23 23:20:08 CET 2003

from __future__ import generators

import sys, os, errno, locale
from random import Random
from codecs import getwriter, getdecoder
from xml.dom.ext.reader import Sax2
from xml.dom.ext import PrettyPrint
from time import time, localtime, strftime
from re import sub
from math import pow

from zettelconf import SGMLDIR

######################################################################
## DOM helper functions

def getText(nodelist):
    str = ""
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            str = str + node.data
    return str

def appendNL(doc,node):
    node.appendChild(doc.createTextNode("\n"))

def insertNLBefore(doc,node,ref):
    node.insertBefore(doc.createTextNode("\n"),ref)

def sanitiseXML(doc):
    """Remove excessive white-space and strange comments."""
    doc.normalize()
    while doc.lastChild.nodeType==doc.COMMENT_NODE:
        doc.removeChild(doc.lastChild)
    zz=doc.getElementsByTagName("kasten")[0].childNodes
    for z in zz:
        if z.nodeType!=z.TEXT_NODE: continue
        z.data=sub("\n( *\n)+","\n",z.data)

def writeXML(doc,fname):
    f=file(fname+"#","w")
    s=PrettyPrint(doc,f)
    f.close()
    os.rename(fname+"#",fname)

######################################################################
## auxiliary functions

def zettels(doc):
    zz=doc.getElementsByTagName("zettel")
    for i in range(0,zz.length): yield zz[i]

def find_ids(doc):
    res=[]
    for z in zettels(doc):
        res.append(z.getAttribute("id"))
    return  res

def gen_id(ids):
    n=0
    for k in ids:
        if k[0]=="A":
            try:
                x=int(k[1:])
                if x>=n: n=x+1
            except:
                pass
    return "A%05d"%n

def normalise_id(id):
    if isinstance(id,int):
        id="A%05d"%id
    else:
        try:
            nid=int(id)
        except:
            pass
        else:
            if str(nid)==id: id="A%05d"%nid
    return id

def simplify_id(id):
    if id[0]=="A":
        try:
            nid=int(id[1:])
            id=str(nid)
        except:
            pass
    return id

def zettel_by_id(doc,id):
    nid=normalise_id(id)
    for z in zettels(doc):
        if z.getAttribute("id") == nid: return z
    raise verzettelt("unknown ID '"+id+"'")

def gen_uuid(rand):
    """Generate a universally unique ID."""
    import sha
    
    uuid=sha.new()
    uuid.update(str(time()))
    uuid.update(str(rand.uniform(0,1)))
    return uuid.hexdigest()

def short_desc(out,z):
    idstr=simplify_id(z.getAttribute("id"))
    state=z.getAttribute("state")
    if not state: state="active"

    a=z.getElementsByTagName("ctime")
    dt=time()-float(a[0].getAttribute("t"))
    if dt>0:
        activestr="*"
    else:
        activestr=" "

    p=z.getElementsByTagName("priority")
    if p:
        prio=float(p[0].getAttribute("p"))
    else:
        prio=1.0

    title=z.getElementsByTagName("title")[0]
    titlestr=getText(title.childNodes)

    return (dt*pow(2,prio),
            "%8s%s %s  %s"%(idstr,activestr,state[0].upper(),titlestr))

######################################################################
## implementation of class Kasten

class verzettelt(Exception):
    def __init__(self, msg):
        self.msg=msg

    def __str__(self):
        return self.msg

class Kasten:
    def __init__(self, fname):
        l=locale.getlocale(locale.LC_CTYPE)
        if l[1]:
            self.writer=getwriter(l[1])
            self.decoder=getdecoder(l[1])
        else:
            self.writer=getwriter("ascii")
            self.decoder=getdecoder("ascii")

        self.rand = Random()

        self.fname=fname
        reader = Sax2.Reader()
        try:
            self.doc=reader.fromStream("file:"+fname)
            self.modified=False
        except OSError, e:
            if e.errno!=errno.ENOENT: raise e
            self.doc=reader.fromString("""<?xml version="1.0" ?>
<!DOCTYPE kasten PUBLIC "-//Jochen Voss//DTD Zettel 1.0//EN" "file://%s/zettel.dtd">
<kasten>
</kasten>
"""%SGMLDIR)
            self.modified=True
        except:
            raise e

        # TODO: remove this hack once 4DOM is fixed
        try:
            doctype=self.doc.firstChild
            pid=doctype.publicId
            sid=doctype.systemId
            if pid.startswith("file:") and sid.startswith("-//"):
                print "zettel: warning: had to work around 4DOM SYSTEM/PUBLIC confusion"
                doctype._publicId=sid
                doctype._systemId=pid
        except:
            pass

        self.ids=find_ids(self.doc)

    def _d(self,str):
        return self.decoder(str)[0]

    def shut(self):
        if self.modified:
            sanitiseXML(self.doc)
            writeXML(self.doc,self.fname)
        self.fname=None
        self.doc=None
        self.ids=None
        self.modified=False

    def list_all(self):
        out=self.writer(sys.stdout,errors='replace')
        lines=[]
        for z in zettels(self.doc): lines.append(short_desc(out,z))
        lines.sort(lambda a,b: (a[0]<b[0])-(a[0]>b[0]))
        for l in lines: out.write(l[1]+"\n")

    def list_current(self,limit=0):
        out=self.writer(sys.stdout,errors='replace')
        lines=[]
        sum=0.0
        for z in zettels(self.doc):
            s=z.getAttribute("state")
            if not s: s="active"
            if s!="active": continue
            a=z.getElementsByTagName("ctime")
            if not a: continue
            t=float(a[0].getAttribute("t"))
            if t>time(): continue
            l=short_desc(out,z)
            lines.append(l)
            sum+=l[0]
        if limit==0 or len(lines)<=limit:
            res=lines
        else:
            res=[]
            for i in range(limit):
                x=self.rand.uniform(0,sum)
                for j in range(len(lines)):
                    x-=lines[j][0]
                    if x<0: break
                res.append(lines[j])
                sum -= lines[j][0]
                del[lines[j]]
        res.sort(lambda a,b: (a[0]<b[0])-(a[0]>b[0]))
        for l in res: out.write(l[1]+"\n")

    def show_by_id(self,id):
        z=zettel_by_id(self.doc,id)
        out=self.writer(sys.stdout,errors='replace')

        title=z.getElementsByTagName("title")[0]
        titlestr=getText(title.childNodes)
        out.write("Title: "+titlestr+"\n")

        out.write("State: "+z.getAttribute("state")+"\n")

        a=z.getElementsByTagName("ctime")
        if a:
            t=float(a[0].getAttribute("t"))
            timestr=strftime("%Y-%m-%d %H:%M", localtime(t))
            out.write("CTime: "+timestr+"\n")

        p=z.getElementsByTagName("priority")
        if p:
            val=float(p[0].getAttribute("p"))
            out.write("Priority: %g\n"%val)

        keys=z.getElementsByTagName("key")
        for k in keys:
            val=k.getAttribute("val")
            out.write("Key: "+val+"\n")

        refs=z.getElementsByTagName("ref")
        for r in refs:
            l=r.getAttribute("l")
            out.write("Ref: "+l+"\n")

        out.write("Id: "+id+"\n")

        body=z.getElementsByTagName("body")[0]
        bodystr=getText(body.childNodes)
        out.write("\n"+bodystr.lstrip())

    def add(self, zdict):
        z=self.doc.createElement("zettel")
        if "id" in zdict:
            id=zdict["id"]
        else:
            id=gen_id(self.ids)
        z.setAttribute("id",id)
        if "uuid" in zdict:
            uuid=zdict["uuid"]
        else:
            uuid=gen_uuid(self.rand)
        z.setAttribute("uuid",uuid)
        appendNL(self.doc,z)

        title=self.doc.createElement("title")
        title.appendChild(self.doc.createTextNode(self._d(zdict["title"])))
        z.appendChild(title)
        appendNL(self.doc,z)

        ctime=self.doc.createElement("ctime")
        if "ctime" in zdict:
            t=int(zdict["ctime"])
        else:
            t=int(time())
        ctime.setAttribute("t","%d"%t)
        z.appendChild(ctime)
        appendNL(self.doc,z)

        if "priority" in zdict:
            priority=self.doc.createElement("priority")
            priority.setAttribute("p","%g"%float(zdict["priority"]))
            z.appendChild(priority)
            appendNL(self.doc,z)

        if "keys" in zdict:
            for k in zdict["keys"]:
                key=self.doc.createElement("key")
                key.setAttribute("val",self._d(k))
                z.appendChild(key)
                appendNL(self.doc,z)

        if "refs" in zdict:
            for r in zdict["refs"]:
                ref=self.doc.createElement("ref")
                ref.setAttribute("l",r)
                z.appendChild(ref)
                appendNL(self.doc,z)

        body=self.doc.createElement("body")
        appendNL(self.doc,body)
        body.appendChild(self.doc.createTextNode(self._d(zdict["body"])))
        appendNL(self.doc,body)
        z.appendChild(body)
        appendNL(self.doc,z)

        zz=self.doc.getElementsByTagName("kasten")[0]
        zz.appendChild(z)
        appendNL(self.doc,zz)

        self.modified = True

    def remove(self,id):
        z=zettel_by_id(self.doc,id)
        z.parentNode.removeChild(z)
        self.modified = True

    def set_state(self,id,state="active"):
        z=zettel_by_id(self.doc,id)
        z.setAttribute("state",state)
        self.modified = True

    def set_priority(self,id,priority=0.0):
        z=zettel_by_id(self.doc,id)
        pp=z.getElementsByTagName("priority")
        if pp:
            p=pp[0]
        else:
            p=self.doc.createElement("priority")
            ref=z.getElementsByTagName("ctime")[0].nextSibling
            z.insertBefore(p,ref)
            insertNLBefore(self.doc,z,ref)
        p.setAttribute("p",str(priority))
        self.modified = True

    def defer(self,id,dt):
        """defer the entry ID until DT seconds in the future"""
        z=zettel_by_id(self.doc,id)
        a=z.getElementsByTagName("ctime")[0]
        a.setAttribute("t",str(int(time()+dt)))
        self.modified = True
